using System;
using UnityEngine;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.VR;
using UnityEngine.XR;
using System.Collections;
using System.Collections.Generic;
using System.Linq;

namespace Blazedust.CUAM
{

    // Based on ResetVROrientation on how to add buttons to the menu-canvas.
    public class CanvasButtons
    {
        
        public static Canvas canvas = null;

        private bool isDesktopMode = false;

        private CustomUnityAssetManager core;

        private UIDynamicButton btnAppend;
        private UIDynamicButton btnReplace;
        private UIDynamicButton btnRemoveLast;
        private UIDynamicButton btnRemoveAll;

        private List<Transform> createdButtons;

        // private bool buttonsEnabled = false;

        private bool lastButtonsOverallEnabled = false;
        private bool lastButtonsVAMEnabled = false;
        private bool lastButtonsCUAMEnabled = false;
        private string lastButtonsLayout = "";
        private float lastButtonsXPosOffset = 0f;
        private float lastButtonsYPosOffset = 0f;

        private string lastButtonsVRLayout = "";
        private float lastButtonsVRXPosOffset = 0f;
        private float lastButtonsVRYPosOffset = 0f;
        private float lastButtonsVRZPosOffset = 0f;

        private float lastButtonsVRXRotOffset = 0f;
        private float lastButtonsVRYRotOffset = 0f;

        private bool lastButtonStateEnabled = false;

        private int lastRemoveAllLinkedAssetQuickCounter = 0;

        public CanvasButtons(CustomUnityAssetManager c)
        {
            core = c;
            createdButtons = new List<Transform>();
        }

        private bool CheckButtonsLayoutChanged()
        {
            Atom selectedAtom = SuperController.singleton.GetSelectedAtom();
            bool buttonsOverallEnabled = core.LogicEnabled && selectedAtom != null && selectedAtom.type == "Person";
            //public JSONStorableBool quickCanvasVAMPresetsButtons;
            //public JSONStorableBool quickCanvasCUAMButtons;
            //public JSONStorableStringChooser quickCanvasButtonsType;
            //public JSONStorableFloat quickCanvasButtonsXOffset;
            //public JSONStorableFloat quickCanvasButtonsYOffset;
            bool layoutChanged = lastButtonsOverallEnabled != buttonsOverallEnabled ||
                                 lastButtonsVAMEnabled != core.options.quickCanvasVAMPresetsButtons.val ||
                                 lastButtonsCUAMEnabled != core.options.quickCanvasCUAMButtons.val ||
                                 lastButtonsLayout != core.options.quickCanvasButtonsType.val ||
                                 lastButtonsXPosOffset != core.options.quickCanvasButtonsXOffset.val ||
                                 lastButtonsYPosOffset != core.options.quickCanvasButtonsYOffset.val ||
                                 lastButtonsVRLayout != core.options.quickCanvasButtonsVRType.val ||
                                 lastButtonsVRXPosOffset != core.options.quickCanvasButtonsVRXOffset.val ||
                                 lastButtonsVRYPosOffset != core.options.quickCanvasButtonsVRYOffset.val ||
                                 lastButtonsVRZPosOffset != core.options.quickCanvasButtonsVRZOffset.val ||
                                 lastButtonsVRXRotOffset != core.options.quickCanvasButtonsVRXRotOffset.val ||
                                 lastButtonsVRYRotOffset != core.options.quickCanvasButtonsVRYRotOffset.val;

            lastButtonsOverallEnabled = buttonsOverallEnabled;
            
            lastButtonsVAMEnabled = core.options.quickCanvasVAMPresetsButtons.val;
            lastButtonsCUAMEnabled = core.options.quickCanvasCUAMButtons.val;
            
            lastButtonsLayout = core.options.quickCanvasButtonsType.val;
            lastButtonsXPosOffset = core.options.quickCanvasButtonsXOffset.val;
            lastButtonsYPosOffset = core.options.quickCanvasButtonsYOffset.val;

            lastButtonsVRLayout = core.options.quickCanvasButtonsVRType.val;
            lastButtonsVRXPosOffset = core.options.quickCanvasButtonsVRXOffset.val;
            lastButtonsVRYPosOffset = core.options.quickCanvasButtonsVRYOffset.val;
            lastButtonsVRZPosOffset = core.options.quickCanvasButtonsVRZOffset.val;
            lastButtonsVRXRotOffset = core.options.quickCanvasButtonsVRXRotOffset.val;
            lastButtonsVRYRotOffset = core.options.quickCanvasButtonsVRYRotOffset.val;

            return layoutChanged;
        }

        public void UpdateButtonLayout()
        {
            if (CheckButtonsLayoutChanged())
            {
                UpdateButtons();
            }

            // check loading/non-loading and button texts
            if (lastButtonsCUAMEnabled)
            {
                bool buttonStateEnabled = (!SuperController.singleton.isLoading && !core.loading.WaitingForPostDelay());
                if (lastButtonStateEnabled != buttonStateEnabled)
                {
                    if (buttonStateEnabled)
                    {
                        if (btnAppend != null)
                            btnAppend.buttonColor = new Color(0.75f, 0.95f, 0.60f);

                        if (btnReplace != null)
                            btnReplace.buttonColor = new Color(0.5f, 0.9f, 0.9f);

                        if (btnRemoveLast != null)
                            btnRemoveLast.buttonColor = new Color(1.0f, 0.7f, 1.0f);

                        if (btnRemoveAll != null)
                            btnRemoveAll.buttonColor = new Color(1.0f, 0.7f, 0.7f);
                    }
                    else
                    {
                        if (btnAppend != null)
                            btnAppend.buttonColor = new Color(0.6f, 0.5f, 0.5f);

                        if (btnReplace != null)
                            btnReplace.buttonColor = new Color(0.6f, 0.5f, 0.5f);

                        if (btnRemoveLast != null)
                            btnRemoveLast.buttonColor = new Color(0.6f, 0.5f, 0.5f);

                        if (btnRemoveAll != null)
                            btnRemoveAll.buttonColor = new Color(0.6f, 0.5f, 0.5f);
                    }
                    lastButtonStateEnabled = buttonStateEnabled;
                }

                if (lastRemoveAllLinkedAssetQuickCounter != core.removeAllLinkedAssetQuickCounter)
                {
                    if (btnRemoveAll != null)
                    {
                        if (core.removeAllLinkedAssetQuickCounter == 0)
                        {
                            btnRemoveAll.label = "[CUAM]\nRemove All";
                        }
                        else
                        {
                            btnRemoveAll.label = "Forced (" + (CustomUnityAssetManager.FORCE_REMOVE_CLICKS - core.removeAllLinkedAssetQuickCounter).ToString() + ")";
                        }
                    }
                    lastRemoveAllLinkedAssetQuickCounter = core.removeAllLinkedAssetQuickCounter;
                }
            }
        }

        private bool CheckPersonAtomSelectedForCUAMAction()
        {
            if (!lastButtonStateEnabled)
            {
                return false;
            }

            Atom selectedAtom = SuperController.singleton.GetSelectedAtom();
            if (selectedAtom.type == "Person")
            {
                if (core.rootAtom == null || core.rootAtom.uid != selectedAtom.uid)
                {
                    core.SelectRootAtom(selectedAtom.uid);
                }
                return true;
            }
            return false;
        }

        private struct ButtonLayoutInfo
        {
            public float x;
            public float y;
            public bool enabled;

            public ButtonLayoutInfo(float posx, float posy)
            {
                x = posx;
                y = posy;
                enabled = true;
            }
            public ButtonLayoutInfo(bool enabledValue, float posx = 0f, float posy = 0f)
            {
                x = posx;
                y = posy;
                enabled = enabledValue;
            }
        }

        public void UpdateButtons()
        {
            // CreateButtons also clears previously buttons to be able to recreate them.
            CreateButtons(() => {

                btnAppend = null;
                btnReplace = null;
                btnRemoveLast = null;
                btnRemoveAll = null;

                if (!lastButtonsOverallEnabled)
                    return;

                // if core.options.enableCanvasButtons ...
                const float bottomY = -0.243f + 0.065f;
                const float leftX = -0.517f;
                const float buttonHeight = 0.045f;
                const float buttonWidth = 0.115f;

                float posx = leftX + (isDesktopMode ? lastButtonsXPosOffset : lastButtonsVRXPosOffset);
                float posy = bottomY + (isDesktopMode ? lastButtonsYPosOffset : lastButtonsVRYPosOffset);
                float posz = (isDesktopMode ? 0f : -0.01f + lastButtonsVRZPosOffset);

                ButtonLayoutInfo[] buttonLayout = new ButtonLayoutInfo[6];

                switch ((isDesktopMode ? lastButtonsLayout : lastButtonsVRLayout))
                {
                    case Options.QUICK_CANVAS_BUTTON_LAYOUT_STACKED:
                        {
                            buttonLayout[0] = new ButtonLayoutInfo(lastButtonsVAMEnabled, posx + buttonWidth, posy + buttonHeight * 2);
                            buttonLayout[1] = new ButtonLayoutInfo(lastButtonsVAMEnabled, posx, posy + buttonHeight * 2);
                            float cuamButtonsYOffset = lastButtonsVAMEnabled ? 0f : buttonHeight;
                            buttonLayout[2] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx, posy + buttonHeight * 1 + cuamButtonsYOffset);
                            buttonLayout[3] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx, posy + buttonHeight * 0 + cuamButtonsYOffset);
                            buttonLayout[4] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx + buttonWidth, posy + buttonHeight * 1 + cuamButtonsYOffset);
                            buttonLayout[5] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx + buttonWidth, posy + buttonHeight * 0 + cuamButtonsYOffset);
                        }
                        break;
                    case Options.QUICK_CANVAS_BUTTON_LAYOUT_HORIZONTAL:
                        {
                            buttonLayout[0] = new ButtonLayoutInfo(lastButtonsVAMEnabled, posx + buttonWidth * 0, posy);
                            buttonLayout[1] = new ButtonLayoutInfo(lastButtonsVAMEnabled, posx + buttonWidth * 1, posy);
                            float cuamButtonsXOffset = lastButtonsVAMEnabled ? 0f : -buttonWidth * 2;
                            buttonLayout[2] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx + buttonWidth * 2 + cuamButtonsXOffset, posy);
                            buttonLayout[3] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx + buttonWidth * 3 + cuamButtonsXOffset, posy);
                            buttonLayout[4] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx + buttonWidth * 4 + cuamButtonsXOffset, posy);
                            buttonLayout[5] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx + buttonWidth * 5 + cuamButtonsXOffset, posy);
                        }
                        break;
                    case Options.QUICK_CANVAS_BUTTON_LAYOUT_VERTICAL:
                        {
                            buttonLayout[0] = new ButtonLayoutInfo(lastButtonsVAMEnabled, posx, posy + buttonHeight * 5);
                            buttonLayout[1] = new ButtonLayoutInfo(lastButtonsVAMEnabled, posx, posy + buttonHeight * 4);
                            float cuamButtonsYOffset = lastButtonsVAMEnabled ? 0f : buttonHeight * 2;
                            buttonLayout[2] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx, posy + buttonHeight * 3 + cuamButtonsYOffset);
                            buttonLayout[3] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx, posy + buttonHeight * 2 + cuamButtonsYOffset);
                            buttonLayout[4] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx, posy + buttonHeight * 1 + cuamButtonsYOffset);
                            buttonLayout[5] = new ButtonLayoutInfo(lastButtonsCUAMEnabled, posx, posy + buttonHeight * 0 + cuamButtonsYOffset);
                        }
                        break;

                }

                ButtonLayoutInfo buttonLayoutInfo = buttonLayout[0];
                if (buttonLayoutInfo.enabled)
                {
                    UIDynamicButton btnAppearance = AddCanvasButton("Select\nAppearance", () =>
                    {
                        try
                        {
                            core.QuickLoadAppearancePreset(SuperController.singleton.GetSelectedAtom());
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }, buttonLayoutInfo.x, buttonLayoutInfo.y, posz, 12f);
                    btnAppearance.textColor = new Color(1.0f, 1.0f, 1.0f);
                    btnAppearance.buttonColor = new Color((40f / 255f), (95f / 255f), (95f / 255f));
                }

                buttonLayoutInfo = buttonLayout[1];
                if (buttonLayoutInfo.enabled)
                {
                    UIDynamicButton btnClothing = AddCanvasButton("Select\nClothing", () =>
                    {
                        try
                        {
                            core.QuickLoadClothingPreset(SuperController.singleton.GetSelectedAtom());
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }, buttonLayoutInfo.x, buttonLayoutInfo.y, posz, 12f);
                    btnClothing.textColor = new Color(1.0f, 1.0f, 1.0f);
                    btnClothing.buttonColor = new Color((152f / 255f), (12f / 255f), (99f / 255f));
                }

                buttonLayoutInfo = buttonLayout[2];
                if (buttonLayoutInfo.enabled)
                {
                    btnAppend = AddCanvasButton("[CUAM]\nAppend", () =>
                    {
                        try
                        {
                            if (CheckPersonAtomSelectedForCUAMAction())
                            {
                                core.QuickLoadAppendPreset(true);
                            }
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }, buttonLayoutInfo.x, buttonLayoutInfo.y, posz, 12f);
                    btnAppend.textColor = new Color(0.0f, 0.0f, 0.0f);
                    btnAppend.buttonColor = new Color(0.75f, 0.95f, 0.60f);
                }

                buttonLayoutInfo = buttonLayout[3];
                if (buttonLayoutInfo.enabled)
                {
                    btnReplace = AddCanvasButton("[CUAM]\nReplace", () =>
                    {
                        try
                        {
                            if (CheckPersonAtomSelectedForCUAMAction())
                            {
                                core.QuickLoadReplacePreset(true);
                            }
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }, buttonLayoutInfo.x, buttonLayoutInfo.y, posz, 12f);
                    btnReplace.textColor = new Color(0.0f, 0.0f, 0.0f);
                    btnReplace.buttonColor = new Color(0.5f, 0.9f, 0.9f);
                }

                buttonLayoutInfo = buttonLayout[4];
                if (buttonLayoutInfo.enabled)
                {
                    btnRemoveLast = AddCanvasButton("[CUAM]\nRemove Last", () =>
                    {
                        try
                        {
                            if (CheckPersonAtomSelectedForCUAMAction())
                            {
                                core.QuickRemoveLastAddedPreset();
                            }
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }, buttonLayoutInfo.x, buttonLayoutInfo.y, posz, 12f);
                    btnRemoveLast.textColor = new Color(0.0f, 0.0f, 0.0f);
                    btnRemoveLast.buttonColor = new Color(1.0f, 0.7f, 1.0f);
                }

                buttonLayoutInfo = buttonLayout[5];
                if (buttonLayoutInfo.enabled)
                {
                    btnRemoveAll = AddCanvasButton("[CUAM]\nRemove All", () =>
                    {
                        try
                        {
                            if (CheckPersonAtomSelectedForCUAMAction())
                            {
                                core.QuickRemoveAllLinkedAssets();
                            }
                        }
                        catch (Exception e)
                        {
                            SuperController.LogError(e.ToString());
                        }
                    }, buttonLayoutInfo.x, buttonLayoutInfo.y, posz, 12f);
                    btnRemoveAll.textColor = new Color(0.0f, 0.0f, 0.0f);
                    btnRemoveAll.buttonColor = new Color(1.0f, 0.7f, 0.7f);
                }

            });
        }

        public void CreateButtons(Action createCanvasButtons)
        {
            float worldScale = SuperController.singleton.worldScale;
            SuperController.singleton.worldScale = 1.0f;

            try
            {
                const float scale = 0.001f;

                Cleanup();
                
                isDesktopMode = !(SuperController.singleton.isOVR || SuperController.singleton.isOpenVR);

                GameObject canvasObject = new GameObject();
                canvas = canvasObject.AddComponent<Canvas>();
                canvas.renderMode = RenderMode.WorldSpace;
                canvas.pixelPerfect = false;
                SuperController.singleton.AddCanvas(canvas);

                canvasObject.layer = SuperController.singleton.mainHUD.gameObject.layer;
                canvas.transform.SetParent(SuperController.singleton.mainHUD, false);

                CanvasScaler cs = canvasObject.AddComponent<CanvasScaler>();
                cs.scaleFactor = 100.0f;
                cs.dynamicPixelsPerUnit = 1f;

                GraphicRaycaster gr = canvasObject.AddComponent<GraphicRaycaster>();

                canvas.transform.localScale = new Vector3(scale, scale, scale);
                canvas.transform.localPosition = new Vector3(0f, 0f, 0.0f);

                LookAtCamera();

                createCanvasButtons.Invoke();

                canvas.transform.Translate(0, 0, 0);
            }
            catch (Exception e)
            {
                SuperController.LogError("CreateButtons(): " + e);
                // DestroyImmediate(this);
            } 
            finally
            {
                SuperController.singleton.worldScale = worldScale;
            }
        }

        public UIDynamicButton AddCanvasButton(string name, UnityAction callback, float x, float y, float z, float width = -40, float height = 42)
        {
            // float z = (isDesktopMode ? 0f : -0.01f);
            Color accessButtonColor = new Color(1.0f, 0.4392f, 0.4392f);
            Color accessTextColor = new Color(0.0f, 0.0f, 0.0f);

            UIDynamicButton button = CreateCanvasButton(name, width, height);
            button.button.onClick.AddListener(callback);
            button.transform.Translate(x, y, z, Space.Self);
            ColorButton(button, accessTextColor, accessButtonColor);

            return button;
        }

        private UIDynamicButton CreateCanvasButton(string name, float width, float height)
        {
            Transform button = GameObject.Instantiate<Transform>(core.manager.configurableButtonPrefab);
            ConfigureTransform(button, width, height);
            ParentToCanvas(button);

            createdButtons.Add(button);

            UIDynamicButton uiButton = button.GetComponent<UIDynamicButton>();
            uiButton.label = name;
            uiButton.buttonText.fontSize = 18;
            return uiButton;
        }

        public static void ColorButton(UIDynamicButton button, Color textColor, Color buttonColor)
        {
            button.textColor = textColor;
            button.buttonColor = buttonColor;
        }

        private void ConfigureTransform(Transform t, float width, float height)
        {
            t.transform.position = Vector3.zero;
            RectTransform rt = t.GetComponent<RectTransform>();
            rt.anchoredPosition = new Vector2(width / 2, height / 2);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Horizontal, width);
            rt.SetSizeWithCurrentAnchors(RectTransform.Axis.Vertical, height);
        }

        private void ParentToCanvas(Transform t)
        {
            t.SetParent(canvas.transform, false);
        }

        private void LookAtCamera()
        {
            if (isDesktopMode)
            {
                canvas.transform.localEulerAngles = new Vector3(0, 180, 0);
            }
            else
            {
                if (XRSettings.enabled == false)
                {
                    Transform cameraT = SuperController.singleton.lookCamera.transform;
                    Vector3 endPos = cameraT.position + cameraT.forward * 10000000.0f;
                    canvas.transform.LookAt(endPos, cameraT.up);
                }
                else
                {
                    canvas.transform.localEulerAngles = new Vector3(28 + lastButtonsVRXRotOffset, 180 + lastButtonsVRYRotOffset, 0);
                }
            }
        }

        public void Cleanup()
        {
            try
            {
                if (createdButtons != null && createdButtons.Count > 0)
                {
                    foreach(Transform t in createdButtons)
                    {
                        if (t.gameObject != null)
                        {
                            GameObject.Destroy(t.gameObject);
                        }
                    }
                    createdButtons.Clear();
                }
                if (canvas != null)
                {
                    if (SuperController.singleton != null)
                    {
                        SuperController.singleton.RemoveCanvas(canvas);
                    }

                    canvas.transform.SetParent(null, false);

                    if (canvas.gameObject != null)
                    {
                        GameObject.Destroy(canvas.gameObject);
                    }
                }
            }
            catch (Exception e)
            {
                SuperController.LogError("Cleanup(): " + e);
                // DestroyImmediate(this);
            }
        }

        public void Destroy()
        {
            Cleanup();
        }
    }
}